#include "autowalker.hpp"
#include "inventory.hpp"

template<typename Comp>
Pathing compute_paths(gui::FSM& fsm, int& count_out) {
  auto& walls_original = fsm.$level.map->$walls;
  auto walls_copy = walls_original;

  Pathing paths{matrix::width(walls_copy), matrix::height(walls_copy)};
  count_out = 0;

  fsm.$level.world->query<components::Position>(
  [&](const auto ent, auto& position) {
      if(ent != fsm.$level.player) {
        if(fsm.$level.world->has<Comp>(ent)) {
          paths.set_target(position.location);
          count_out = count_out + 1;
        } else {
          // this will mark that spot as a wall so we don't path there temporarily
          walls_copy[position.location.y][position.location.x] = WALL_PATH_LIMIT;
        }
      }
  });

  paths.compute_paths(walls_copy);

  return paths;
}

Pathing Autowalker::path_to_enemies() {
  return compute_paths<components::Combat>(fsm, enemy_count);
}

Pathing Autowalker::path_to_items() {
  return compute_paths<components::InventoryItem>(fsm, item_count);
}

Pathing Autowalker::path_to_devices() {
  return compute_paths<components::Device>(fsm, device_count);
}

void Autowalker::window_events() {
  fsm.$window.handleEvents(
      [&](const sf::Event::KeyPressed &) {
        fsm.autowalking = false;
        fmt::println("ABORT AUTOWALK");
      },
      [&](const sf::Event::MouseButtonPressed &) {
        fsm.autowalking = false;
        fmt::println("ABORT AUTOWALK");
      }
  );
}

void Autowalker::process_combat() {
  while(fsm.in_state(gui::State::IN_COMBAT)
      || fsm.in_state(gui::State::ATTACKING))
  {
    if(fsm.in_state(gui::State::ATTACKING)) {
      send_event(gui::Event::TICK);
    } else {
      send_event(gui::Event::ATTACK);
    }
  }
}

Point Autowalker::get_current_position() {
  auto& player_position = fsm.$level.world->get<components::Position>(fsm.$level.player);
  return player_position.location;
}

bool Autowalker::path_player(Pathing& paths, Point& target_out) {
  bool found = paths.random_walk(target_out, false, PATHING_TOWARD);

  if(!found) {
    dbc::log("no neighbor found, aborting autowalk");
    return false;
  }

  if(!fsm.$level.map->can_move(target_out)) {
    dbc::log("neighbors is telling me to go to a bad spot.");
    return false;
  }

  return true;
}

void Autowalker::rotate_player(Point current, Point target) {
  int delta_x = int(target.x) - int(current.x);
  int delta_y = int(target.y) - int(current.y);

  int facing = fsm.$main_ui.$compass_dir;
  int target_facing = 0;

  if(delta_x == -1 && delta_y == 0) {
    // west
    target_facing = 4;
  } else if(delta_x == 1 && delta_y == 0) {
    // east
    target_facing = 0;
  } else if(delta_x == 0 && delta_y == 1) {
    // south
    target_facing = 2;
  } else if(delta_x == 0 && delta_y == -1) {
    // north
    target_facing = 6;
  } else {
    dbc::sentinel(
        fmt::format("got more than 4 direction result: "
          "current={},{} "
          "target={},{} "
          "delta={},{} ",
          current.x, current.y,
          target.x, target.y,
          delta_x, delta_y));
  }

  auto dir = facing > target_facing ? gui::Event::ROTATE_LEFT : gui::Event::ROTATE_RIGHT;

  while(facing != target_facing) {
    send_event(dir);
    facing = fsm.$main_ui.$compass_dir;
  }

  dbc::check(fsm.$main_ui.$compass_dir == target_facing,
      "player isn't facing the correct direction");
}

void Autowalker::show_map_overlay(matrix::Matrix& map, Point current) {
  auto debug = fsm.$level.world->get_the<components::Debug>();
  if(!debug.FPS) {
    fsm.$main_ui.$overlay_ui.close_text("top_right");
    return;
  }

  std::string map_overlay;
  for(matrix::box it{map, current.x, current.y, 6}; it.next();) {
      if(it.x == it.left) map_overlay += "\n";
      int cell = map[it.y][it.x];

      if(it.x == current.x && it.y == current.y) {
        map_overlay += fmt::format("{:x}<", cell);
      } else if(cell == WALL_PATH_LIMIT) {
        map_overlay += fmt::format("# ");
      } else if(cell > 15) {
        map_overlay += fmt::format("* ");
      } else {
        map_overlay += fmt::format("{:x} ", cell);
      }
  }

  fsm.$main_ui.$overlay_ui.show_text("top_right", map_overlay);
}

void Autowalker::autowalk() {
  window_events();
  if(!fsm.autowalking) return;

  process_combat();
  auto paths = path_to_enemies();

  if(enemy_count == 0) {
    dbc::log("Killed everything, now finding items.");
    paths = path_to_items();
  }

  if(enemy_count == 0 && item_count == 0) {
    dbc::log("No more items, find the exit.");
    paths = path_to_devices();
  }

  if(enemy_count == 0 &&
      item_count == 0 &&
      device_count == 0)
  {
    fsm.autowalking = false;
    dbc::log("no more enemies, items, or devices.");
    return;
  }

  Point current = get_current_position();
  Point target = current;

  show_map_overlay(paths.$paths, current);


  if(!path_player(paths, target)) {
    dbc::log("no paths found, aborting autowalk");
    fsm.autowalking = false;
    return;
  }

  rotate_player(current, target);

  int move_attempts = 0;
  do {
    process_combat();
    process_move();
    // BUG: sometimes in idle but there's an enemy near but combat hasn't started
    // for now just toss out an ATTACK and it'll be ignored or cause combat
    send_event(gui::Event::ATTACK);
    move_attempts++;
  } while(move_attempts < 100 && !player_has_moved(target));
}

void Autowalker::process_move() {
  send_event(gui::Event::MOVE_FORWARD);
  while(fsm.in_state(gui::State::MOVING)) send_event(gui::Event::TICK);
}

bool Autowalker::player_has_moved(Point target) {
  Point current = get_current_position();
  return current.x == target.x && current.y == target.y;
}

void Autowalker::send_event(gui::Event ev) {
  fsm.event(ev);
  fsm.render();
  fsm.handle_world_events();
}

void Autowalker::start_autowalk() {
  fsm.autowalking = true;
}
